/* afb_lib.c   */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <bios.h>
#include <conio.h>
#include <at200.h>
#include <camera.h>

#include "idecls.h"
#include "camsys.h"
#include "proto.h"
#include "serial.h"
#include "mode.h"

/* Local defines */

#define T_TEI	1
#define T_TEE	2
#define T_TD	3
/*			DELETED
#define T_TCC	4
*/

#define T_TCF	5
#define T_TFW	6

static int sel_port = -1;		/* currently selected COM port */

extern AFB_CONST afb_const[2];		/* Defined in KI.C */
extern int comport;

/*------------ Auxiliary Function Board routines.  -------------------------*/

pv_ptr sel ( register pv_ptr p )    /* Select camera, make it current. */
{
     int  typ, camera_no, sav_cam;

     sav_cam = sel_cam;			/* In case of an error. */
     p = clause ( p, &typ, &camera_no );
     if ( typ != CONST )
     {
     	sprintf( omsg, "ERROR: 'cam' value expected.\n" );
     	disp_omsg( ERROR );
     	sel_cam = sav_cam;
     	return ( p );
     }

     if ( ( camera_no == 0 ) || ( camera_no == 1 ) )
     {
         select_camera( camera_no );
     }
     else
     {
         sprintf( omsg,"ERROR: camera %d does not exist \n", sel_cam);
         disp_omsg( ERROR );
         sel_cam = sav_cam;
     } 
     return ( p );
}

pv_ptr sfw ( register pv_ptr p )    /* Set Filter Wheel Position */
{
     int  typ, pos;
     int  indata;

     char setfw = 0x20;

     /* Pick up command arg: 'pos' */
     
     p = clause ( p, &typ, &pos );
     if ( typ != CONST )
     {
     	sprintf( omsg, "ERROR: 'pos' value expected - range 0..4\n" );
     	disp_omsg( ERROR );
     	return ( p );
     }     

     if ( ( pos < 0 ) || ( pos > 4 ) ) 
     {
          sprintf( omsg,"ERROR: Filter Wheel Number out of range 0..4\n");
          disp_omsg( ERROR );
          return ( p );
     }
     
     /* Send command to set new position. */

     setfw = (char) ( ( ( setfw | ( sel_cam << 7 ) ) | pos ) & 0x00FF ); 

     /* Send inquiry, test for handshake error */
     
     if ( com_send( sel_misc->portno, setfw ) )	
          return ( p );					/* failed */
     
     /* Wait for reply from AFB. */ 

     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        indata &= 0x07;			/* Mask off hi-order bits */
        if ( indata != pos )
        {
           sprintf( omsg,"ERROR: Filter Wheel Not Set to the Correct Position.\n");
           disp_omsg( ERROR );
           return ( p );
        }
        else
        {
           sprintf( omsg, "cam %d: fw	%d \n", sel_cam, indata );
           disp_omsg( NOERROR );
           sel_misc->fw = (unsigned char) indata; 

           /* Update filter wheel description */ 
   
           strcpy ( sel_misc->filter_desc,  filter_desc[ sel_cam ][ indata ] );
   
        }
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr afb_id ( register pv_ptr p )    /* Read AFB ID Bits */
{
     int  indata, i, x;
     char inq = 0x01;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        indata &= 0x07;			/* Mask off hi-order bits */

	/* See if indata matches any id in afb_const table. */

	for ( i = 0, x = -1; i < 2; i++ )
	{
		if ( indata == (int) afb_const[ i ].id )
		{
			x = i;
			break;
		}
	}
	if ( x == -1 )		/* No match found. */
	{
		sprintf( omsg, "** No match on AFB ID for camera %d\n", sel_cam );
		disp_omsg( ERROR );
	}
	else
	{
		sprintf( omsg,"cam %d  AFB ID:	%d\n", sel_cam, indata );
		disp_omsg( NOERROR );
		
		/* Store new value of ID for AFB */
	
        	sel_misc->afb_id = (unsigned char) indata; 	
        }
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr rfw ( register pv_ptr p )    /* Read Filter Wheel Position */
{
     int  indata;
     char inq = 0x02;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        indata &= 0x07;			/* Mask off hi-order bits */
        sprintf( omsg,"cam %d: fw	%d\n", sel_cam, indata );
	disp_omsg( NOERROR );
        sel_misc->fw = (unsigned char) indata; 
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr tei ( register pv_ptr p )    /* Read Enclosure Internal Ambient Temp. */
{
     int  indata;
     char inq = 0x03;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        update_temp( sel_misc->temp.tei, indata, T_TEI );
        sprintf( omsg,"cam %d: tei	%d\n", sel_cam, sel_misc->temp.tei[0] );
	disp_omsg( NOERROR );
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr tee ( register pv_ptr p )    /* Read Enclosure External Ambient Temp. */
{
     int  indata;
     char inq = 0x04;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        update_temp( sel_misc->temp.tee, indata, T_TEE );
        sprintf( omsg, "cam %d: tee	%d\n", sel_cam, sel_misc->temp.tee[0] );
	disp_omsg( NOERROR );
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr td ( register pv_ptr p )    /* Read Dome Temperature */
{
     int  indata;
     char inq = 0x05;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        update_temp( sel_misc->temp.td, indata, T_TD );
	sprintf( omsg, "cam %d: td	%d\n", sel_cam, sel_misc->temp.td[0] );
	disp_omsg( NOERROR );
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr tcf ( register pv_ptr p )    /* Read CCD Temperature, Fine */
{
     int  indata;
     char inq = 0x07;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        update_temp( sel_misc->temp.tcf, indata, T_TCF );
        sprintf( omsg, "cam %d: tcf	%d\n", sel_cam, sel_misc->temp.tcf[0] );
	disp_omsg( NOERROR );
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr tfw ( register pv_ptr p )    /* Read Filter Wheel Temperature */
{
     int  indata;
     char inq = 0x08;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        update_temp( sel_misc->temp.tfw, indata, T_TFW );
        sprintf( omsg, "cam %d: tfw	%d\n", sel_cam, sel_misc->temp.tfw[0] );
	disp_omsg( NOERROR );
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr heater ( register pv_ptr p )    /* Read Heater Status */
{
     int  indata;
     char inq = 0x09;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        sprintf( omsg, "cam %d Heater: %d \n", sel_cam, indata );
	disp_omsg( NOERROR );
        sel_misc->heater = indata; 
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr acpower ( register pv_ptr p )    /* Read power - AC to electronics. */
{
     int  indata;
     char inq = 0x0A;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        sprintf( omsg,"cam %d AC power: %d \n",sel_cam, indata );
	disp_omsg( NOERROR );
        sel_misc->acpower = indata; 
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr plus5 ( register pv_ptr p )    /* Read power - plus 5 volt rail. */
{
     int  indata, x, result;
     char inq = 0x0B;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
	/* Match AFB id with constants table entry. */

	if ( ( x = match_afb() ) == -1 )		/* No match found. */
	{
		sprintf( omsg, "** No match on AFB ID for cam %d\n", sel_cam );
		disp_omsg( ERROR );
		result = 0;
	}
	else
	{
           result = ( indata * afb_const[x].plus5 ) / 256;	
           sprintf(omsg,"cam %d +5 V status: %d\n", sel_cam, result);
	   disp_omsg( NOERROR );
           sel_misc->plus5 = result; 
	}
     }
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr plus15 ( register pv_ptr p )    /* Read power - plus 15 volt rail. */
{
     int  indata, x, result;
     char inq = 0x0C;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
	/* Match AFB id with constants table entry. */

	if ( ( x = match_afb() ) == -1 )		/* No match found. */
	{
		sprintf( omsg, "** No match on AFB ID for cam %d\n", sel_cam );
		disp_omsg( ERROR );
		result = 0;
	}
	else
	{
		result = ( indata * afb_const[x].plus15 ) / 256;	
        	sprintf( omsg,"cam %d +15 V status: %d\n", sel_cam, result );
		disp_omsg( NOERROR );
        	sel_misc->plus15 = result;
        } 
     }

     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr minus15 ( register pv_ptr p )    /* Read power - minus 15 volt rail. */
{
     int  indata, x, result;
     char inq = 0x0D;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
	/* Match AFB id with constants table entry. */

	if ( ( x = match_afb() ) == -1 )		/* No match found. */
	{
		sprintf( omsg, "** No match on AFB ID for cam %d\n", sel_cam );
		disp_omsg( ERROR );
		result = 0;
	}
	else
	{
		result = ( indata * afb_const[x].minus15 ) / 256;	
	        sprintf( omsg,"cam %d -15 V status: %d\n", sel_cam, result );
		disp_omsg( NOERROR );
	        sel_misc->minus15 = result;
	} 
     }

     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr fan_stat ( register pv_ptr p )    /* Read Fan Status */
{
     int  indata;
     char inq = 0x0E;

     inq = (char) ( ( inq | sel_cam << 7 ) & 0x00FF );

     if ( com_send( sel_misc->portno, inq ) )
               return ( p );
     
     if ( com_rcv( sel_misc->portno, &indata ) )
     {
        sprintf( omsg, "cam %d Fan Status: %d \n", sel_cam, indata );
	disp_omsg( NOERROR );
        sel_misc->fan_stat = indata; 
     }
             
     return ( p );
}

/*----------------------------------------------------------------------*/

pv_ptr cam_stat ( register pv_ptr p )    /* Show camera status */
{
    sprintf( omsg, "\nMISC fields:             cam %d, controller %X status:\n\n", 
              sel_cam, at200_base ); 
    disp_omsg( NOERROR );

    sprintf( omsg, "AFB id      = %4d       ", sel_misc->afb_id ); 
    disp_omsg( NOERROR );
    sprintf( omsg, "filter wheel= %4d       ", sel_misc->fw ); 
    disp_omsg( NOERROR );
    sprintf( omsg, "ems_handle  = %4d\n",      sel_misc->ems_handle );
    disp_omsg( NOERROR );
    sprintf( omsg, "exp_type    =    %c       ", sel_misc->exposure_type );
    disp_omsg( NOERROR );
    sprintf( omsg, "exp_scale   =%5u       ",  sel_misc->exp_scale );
    disp_omsg( NOERROR );
    sprintf( omsg, "preamp      = %4d\n",      sel_misc->preamp ); 
    disp_omsg( NOERROR );
/*  sprintf( omsg, ":%s\n",                    sel_misc->filter_desc ); 
    disp_omsg( NOERROR );                                               */
    sprintf( omsg, "tfw         = %4d       ", sel_misc->temp.tfw[0] );
    disp_omsg( NOERROR );
    sprintf( omsg, "heater      = %4d       ", sel_misc->heater );
    disp_omsg( NOERROR );
    sprintf( omsg, "fan         = %4d\n",      sel_misc->fan_stat );
    disp_omsg( NOERROR );
    sprintf( omsg, "tei         = %4d       ", sel_misc->temp.tei[0] );
    disp_omsg( NOERROR );
    sprintf( omsg, "ACpower     = %4d\n",      sel_misc->acpower );
    disp_omsg( NOERROR );
    sprintf( omsg, "td          = %4d       ", sel_misc->temp.td[0] );
    disp_omsg( NOERROR );
    sprintf( omsg, "plus5       = %4d\n",      sel_misc->plus5 );
    disp_omsg( NOERROR );
    sprintf( omsg, "tee         = %4d       ", sel_misc->temp.tee[0] );
    disp_omsg( NOERROR );
    sprintf( omsg, "plus15      = %4d\n",      sel_misc->plus15 );
    disp_omsg( NOERROR );
    sprintf( omsg, "tcf         = %4d       ", sel_misc->temp.tcf[0] );
    disp_omsg( NOERROR );
    sprintf( omsg, "minus15     = %4d\n"     , sel_misc->minus15 );
    disp_omsg( NOERROR );
    sprintf( omsg, "pixel_min   =%5d       " , sel_misc->pixel_min );
    disp_omsg( NOERROR );
    sprintf( omsg, "pixel_max   =%5d\n\n"    , sel_misc->pixel_max );
    disp_omsg( NOERROR );

    sprintf( omsg, "CAM_EXP fields          " );
    disp_omsg( NOERROR );
    sprintf( omsg, "CAM_SYS fields\n" );
    disp_omsg( NOERROR );
    sprintf( omsg, "exposure    =%5ld      ",  sel_exp->exposure );
    disp_omsg( NOERROR );
    sprintf( omsg, "at200_id    = %4d\n",      sel_sys->at200_id );
    disp_omsg( NOERROR );
    sprintf( omsg, "gain        = %4d      ",  sel_exp->gain ? 4 : 1 );
    disp_omsg( NOERROR );
    sprintf( omsg, "frame       = %4d\n",      sel_sys->frame );
    disp_omsg( NOERROR );
    sprintf( omsg, "corg        = %4d      ",  sel_exp->corg );
    disp_omsg( NOERROR );
    sprintf( omsg, "open_delay  = %4d\n",      sel_sys->open_delay );
    disp_omsg( NOERROR );
    sprintf( omsg, "rorg        = %4d      ",  sel_exp->rorg );
    disp_omsg( NOERROR );
    sprintf( omsg, "close_delay = %4d\n",      sel_sys->close_delay );
    disp_omsg( NOERROR );
    sprintf( omsg, "cols        = %4d      ",  sel_exp->cols );
    disp_omsg( NOERROR );
    sprintf( omsg, "preamp_delay= %4d\n",      sel_sys->preamp_delay );
    disp_omsg( NOERROR );
    sprintf( omsg, "rows        = %4d      ",  sel_exp->rows );
    disp_omsg( NOERROR );
    sprintf( omsg, "speed       = %4d\n",      sel_sys->speed );
    disp_omsg( NOERROR );
    sprintf( omsg, "sbin        = %4d      ",  sel_exp->sbin );
    disp_omsg( NOERROR );
    sprintf( omsg, "data_bits   = %4d\n",      sel_sys->data_bits );
    disp_omsg( NOERROR );

    sprintf( omsg, "pbin        = %4d\n",      sel_exp->pbin );
    disp_omsg( NOERROR );
    sprintf( omsg, "mpp_mode    = %4d\n",      sel_exp->mpp_mode );
    disp_omsg( NOERROR );
    sprintf( omsg, "clr_cnt     = %4d\n",      sel_exp->clr_cnt );
    disp_omsg( NOERROR );
    sprintf( omsg, "row_shift   = %4d\n\n",    sel_exp->row_shift );
    disp_omsg( NOERROR );

    return ( p );
    
}

/*----------------------------------------------------------------------*/

pv_ptr reset_afb ( register pv_ptr p )    /* Reset AFB for selected camera. */
{
     int  indata;
     char rs_cod = 0x73;		/* Reset code. */

     /* Send reset command. */

     rs_cod = (char) ( ( rs_cod | ( sel_cam << 7 ) ) & 0x00FF ); 

     if ( com_send( sel_misc->portno, rs_cod ) )
          return ( p );
     
     /* Wait for reply from AFB. */ 

     com_rcv( sel_misc->portno, &indata );

     if ( indata != 0xA5 )
     {
           sprintf( omsg,"(reset_afb) ERROR: AFB %d on COM%d did not return correct 'reset' code.\n",
           		sel_misc->afb_id, sel_misc->portno );
           disp_omsg( ERROR );
           return ( p );
     }
     return ( p );
}


/* Send inquiry character out on serial port.
 * Returns FALSE if no error, TRUE if exception.
 */

int com_send( int port, char inq )
{
int 	dummy;

	/* Clear any characters in input buffer. */

	while ( ( dummy = s_rcvchar() ) != -1 );
	
	return !s_sendchar( inq );	/* Then send our character. */
}

/* Receive char from serial port, check status. 
 * Data is returned in rcv_data.
 * Returns TRUE if no error, FALSE if exception.
 */

int com_rcv( int port, int *rcv_data )
{
     unsigned int  kbin, com_in;
     int  DONE = FALSE;
     int  ESC  = FALSE;
     unsigned int  timeout = 0x2000;

     while ( !DONE && !ESC ) 
     {
     	/* See if serial port has a data character to read. */

	if ( ( com_in = s_rcvchar() ) != -1 )
	{
		*rcv_data = com_in & 0xff;	/* strip status byte off */
        	DONE = TRUE;
	}
	else
	{
		/* No - see if user pressed escape key. */	

		if ( kbhit() ){
        	   if( ( kbin = getch() ) == '\x1B' ) 
        	      ESC = TRUE;
		   else ungetch( kbin );
		}
		
		/* See if timed out without ever seeing DATA_READY. */

		if ( !( --timeout ) )
		{
		   	sprintf( omsg,"(com_rcv) COM%d timeout.\n", port );

	   		disp_omsg( ERROR );
	  		return FALSE;
		}
	}
     }
     return TRUE;
}
             
/* Select a camera by setting all camera-specific addresses. */

void select_camera ( int camera_number )
{
int int_mask, int_enable_mask, int_disable_mask;

     sel_cam = camera_number;
     if ( camera_number == 0 )
     {
     	if ( std0_misc.cam_opmode & 0x1 )
     	{
    		sel_sys  = &std0_sys;	/* Select camera 0 */
    		sel_exp  = &std0_exp;
    		sel_misc = &std0_misc;
    		sel_hdr  = &header0;
        	at200_base = 0x220;
             	
        	int_enable_mask = 0xEF;
        	int_disable_mask = 0x08;
	}
	else
	{
		sprintf( omsg,"ERROR: Camera 0 not enabled.\n" );
		disp_omsg( ERROR );
	}
     }
     else if (camera_number == 1 )
     {
     	if ( std1_misc.cam_opmode & 0x2 )
     	{
    		sel_sys  = &std1_sys;	/* Select camera 1 */
    		sel_exp  = &std1_exp;
    		sel_misc = &std1_misc;
    		sel_hdr  = &header1;
        	at200_base = 0x230;
             	
        	int_enable_mask = 0xF7;
        	int_disable_mask = 0x10;
	}
	else
	{
		sprintf( omsg,"ERROR: Camera 1 not enabled.\n" );
		disp_omsg( ERROR );
	}
     }
     comport = *( BIOS_DATA + camera_number );
     int_mask = ( (inp( P8259_1) & int_enable_mask ) | int_disable_mask );
     exec_compress = do_compress[camera_number];
          
     _disable();
     outp( MCR, MCRALL );
     outp( IER, IERALL );
     outp( P8259_1, int_mask );
     _enable();
}

/* Store current temperature reading, set new min, max, mean, and 
 * number of readings. (temperature readings are stored X10 actual value
 * thus providing nnn.m values in integer form)
 */
 
void update_temp( int *temp_array, int data_in, int which_temp )
{
	int	x;
	long    lo, hi;
	long	result;
	
	/* Match AFB id with constants table entry. */

	if ( ( x = match_afb() ) == -1 )		/* No match found. */
	{
		sprintf( omsg, "** No match on AFB ID for camera # %d\n", 
				sel_cam );
		disp_omsg( ERROR );
		return;					/* No update done. */
	}
	else
	{
		switch ( which_temp )		/* Get hi/lo temp range */
		{
			case T_TEE:
				hi = afb_const[x].tee[0];
				lo = afb_const[x].tee[1];
				break;
			case T_TEI:
				hi = afb_const[x].tei[0];
				lo = afb_const[x].tei[1];
				break;
			case T_TD:
				hi = afb_const[x].td[0];
				lo = afb_const[x].td[1];
				break;
			case T_TCF:
				hi = afb_const[x].tcf[0];
				lo = afb_const[x].tcf[1];
				break;
			case T_TFW:
				hi = afb_const[x].tfw[0];
				lo = afb_const[x].tfw[1];
		}
		
		result = (long) ( ( ((long)data_in) * ( hi - lo ) * 10L ) / 256L ) 
		         + (long) ( lo * 10L );
	}

	temp_array[TEMP_CUR] = (int) result;
	if ( temp_array[TEMP_NO]++ == 0 )		/* First reading */
	{
		temp_array[TEMP_MIN] = temp_array[TEMP_MAX] 
		= temp_array[TEMP_MEAN] = (int) result;
	}
	else
	{
		if ( result < temp_array[TEMP_MIN] ) 
			temp_array[TEMP_MIN] = (int) result;
		if ( result > temp_array[TEMP_MAX] ) 
			temp_array[TEMP_MAX] = (int) result;
		temp_array[TEMP_MEAN] = (temp_array[TEMP_MEAN] 
					+ (int) result) / 2;
	}
}


/* Match AFB id for selected camera with constants table entry. 
 * Returns index to matched AFB constants table entry.
 */

int match_afb( void )
{
	int	i;

	for ( i = 0; i < 2; i++ )
		if ( sel_misc->afb_id == afb_const[ i ].id )
			return i;

	return -1;	/* No match found. */
}
